"""Types for representing symbolic expressions for geometric quantities."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.core.terminal import Terminal
from ufl.core.ufl_type import ufl_type
from ufl.domain import as_domain, extract_unique_domain
from ufl.sobolevspace import H1

"""
Possible coordinate bootstrapping:

Xf = Xf[q]
    FacetCoordinate = quadrature point on facet (ds,dS)

X = X[q]
    CellCoordinate = quadrature point on cell (dx)

x = x[q]
    SpatialCoordinate = quadrature point from input array (dc)


Jacobians of mappings between coordinates:

Jcf = dX/dXf = grad_Xf X(Xf)
    CellFacetJacobian

Jxc = dx/dX = grad_X x(X)
    Jacobian

Jxf = dx/dXf = grad_Xf x(Xf)  =  Jxc Jcf = dx/dX dX/dXf = grad_X x(X) grad_Xf X(Xf)
    FacetJacobian = Jacobian * CellFacetJacobian


Possible computation of X from Xf:

X = Jcf Xf + X0f
    CellCoordinate = CellFacetJacobian * FacetCoordinate + CellFacetOrigin


Possible computation of x from X:

x = f(X)
    SpatialCoordinate = sum_k xdofs_k * xphi_k(X)

x = Jxc X + x0
    SpatialCoordinate = Jacobian * CellCoordinate + CellOrigin


Possible computation of x from Xf:

x = x(X(Xf))

x = Jxf Xf + x0f
    SpatialCoordinate = FacetJacobian * FacetCoordinate + FacetOrigin


Inverse relations:

X = K * (x - x0)
    CellCoordinate = JacobianInverse * (SpatialCoordinate - CellOrigin)

Xf = FK * (x - x0f)
    FacetCoordinate = FacetJacobianInverse * (SpatialCoordinate - FacetOrigin)

Xf = CFK * (X - X0f)
    FacetCoordinate = CellFacetJacobianInverse * (CellCoordinate - CellFacetOrigin)
"""


# --- Expression node types


@ufl_type(is_abstract=True)
class GeometricQuantity(Terminal):
    """Geometric quantity."""

    __slots__ = ("_domain",)

    def __init__(self, domain):
        """Initialise."""
        Terminal.__init__(self)
        self._domain = as_domain(domain)

    def ufl_domains(self):
        """Get the UFL domains."""
        return (self._domain,)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # NB! Geometric quantities are piecewise constant by
        # default. Override if needed.
        return True

    # NB! Geometric quantities are scalar by default. Override if
    # needed.
    ufl_shape = ()

    def _ufl_signature_data_(self, renumbering):
        """Signature data of geometric quantities depend on the domain numbering."""
        return (self._ufl_class_.__name__,) + self._domain._ufl_signature_data_(renumbering)

    def __str__(self):
        """Format as a string."""
        return self._ufl_class_.name

    def __repr__(self):
        """Representation."""
        r = "%s(%s)" % (self._ufl_class_.__name__, repr(self._domain))
        return r

    def _ufl_compute_hash_(self):
        """UFL compute hash."""
        return hash((type(self).__name__,) + self._domain._ufl_hash_data_())

    def __eq__(self, other):
        """Check equality."""
        return isinstance(other, self._ufl_class_) and other._domain == self._domain


@ufl_type(is_abstract=True)
class GeometricCellQuantity(GeometricQuantity):
    """Geometric cell quantity."""

    __slots__ = ()


@ufl_type(is_abstract=True)
class GeometricFacetQuantity(GeometricQuantity):
    """Geometric facet quantity."""

    __slots__ = ()


# --- Coordinate represented in different coordinate systems


@ufl_type()
class SpatialCoordinate(GeometricCellQuantity):
    """The coordinate in a domain.

    In the context of expression integration,
    represents the domain coordinate of each quadrature point.

    In the context of expression evaluation in a point,
    represents the value of that point.
    """

    __slots__ = ()
    name = "x"

    @property
    def ufl_shape(self):
        """Return the number of coordinates defined (i.e. the geometric dimension of the domain)."""
        g = self._domain.geometric_dimension()
        return (g,)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # Only case this is true is if the domain is a vertex cell.
        t = self._domain.topological_dimension()
        return t == 0

    def evaluate(self, x, mapping, component, index_values):
        """Return the value of the coordinate."""
        if component == ():
            if isinstance(x, (tuple, list)):
                return float(x[0])
            else:
                return float(x)
        else:
            return float(x[component[0]])

    def count(self):
        """Count."""
        # FIXME: Hack to make SpatialCoordinate behave like a coefficient.
        # When calling `derivative`, the count is used to sort over.
        return -1


@ufl_type()
class CellCoordinate(GeometricCellQuantity):
    """The coordinate in a reference cell.

    In the context of expression integration,
    represents the reference cell coordinate of each quadrature point.

    In the context of expression evaluation in a point in a cell,
    represents that point in the reference coordinate system of the cell.
    """

    __slots__ = ()
    name = "X"

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        t = self._domain.topological_dimension()
        return (t,)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # Only case this is true is if the domain is a vertex cell.
        t = self._domain.topological_dimension()
        return t == 0


@ufl_type()
class FacetCoordinate(GeometricFacetQuantity):
    """The coordinate in a reference cell of a facet.

    In the context of expression integration over a facet,
    represents the reference facet coordinate of each quadrature point.

    In the context of expression evaluation in a point on a facet,
    represents that point in the reference coordinate system of the facet.
    """

    __slots__ = ()
    name = "Xf"

    def __init__(self, domain):
        """Initialise."""
        GeometricFacetQuantity.__init__(self, domain)
        t = self._domain.topological_dimension()
        if t < 2:
            raise ValueError("FacetCoordinate is only defined for topological dimensions >= 2.")

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        t = self._domain.topological_dimension()
        return (t - 1,)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # Only case this is true is if the domain is an interval cell
        # (with a vertex facet).
        t = self._domain.topological_dimension()
        return t <= 1


# --- Origin of coordinate systems in larger coordinate systems


@ufl_type()
class CellOrigin(GeometricCellQuantity):
    """The spatial coordinate corresponding to origin of a reference cell."""

    __slots__ = ()
    name = "x0"

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        g = self._domain.geometric_dimension()
        return (g,)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        return True


@ufl_type()
class FacetOrigin(GeometricFacetQuantity):
    """The spatial coordinate corresponding to origin of a reference facet."""

    __slots__ = ()
    name = "x0f"

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        g = self._domain.geometric_dimension()
        return (g,)


@ufl_type()
class CellFacetOrigin(GeometricFacetQuantity):
    """The reference cell coordinate corresponding to origin of a reference facet."""

    __slots__ = ()
    name = "X0f"

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        t = self._domain.topological_dimension()
        return (t,)


# --- Jacobians of mappings between coordinate systems


@ufl_type()
class Jacobian(GeometricCellQuantity):
    r"""The Jacobian of the mapping from reference cell to spatial coordinates.

    .. math:: J_{ij} = \\frac{dx_i}{dX_j}
    """

    __slots__ = ()
    name = "J"

    @property
    def ufl_shape(self):
        """Return the number of coordinates defined (i.e. the geometric dimension of the domain)."""
        g = self._domain.geometric_dimension()
        t = self._domain.topological_dimension()
        return (g, t)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # Only true for a piecewise linear coordinate field in simplex cells
        return self._domain.is_piecewise_linear_simplex_domain()


@ufl_type()
class FacetJacobian(GeometricFacetQuantity):
    """The Jacobian of the mapping from reference facet to spatial coordinates.

      FJ_ij = dx_i/dXf_j

    The FacetJacobian is the product of the Jacobian and CellFacetJacobian:

      FJ = dx/dXf = dx/dX dX/dXf = J * CFJ
    """

    __slots__ = ()
    name = "FJ"

    def __init__(self, domain):
        """Initialise."""
        GeometricFacetQuantity.__init__(self, domain)
        t = self._domain.topological_dimension()
        if t < 2:
            raise ValueError("FacetJacobian is only defined for topological dimensions >= 2.")

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        g = self._domain.geometric_dimension()
        t = self._domain.topological_dimension()
        return (g, t - 1)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # Only true for a piecewise linear coordinate field in simplex
        # cells
        return self._domain.is_piecewise_linear_simplex_domain()


@ufl_type()
class CellFacetJacobian(GeometricFacetQuantity):  # dX/dXf
    """The Jacobian of the mapping from reference facet to reference cell coordinates.

    CFJ_ij = dX_i/dXf_j
    """

    __slots__ = ()
    name = "CFJ"

    def __init__(self, domain):
        """Initialise."""
        GeometricFacetQuantity.__init__(self, domain)
        t = self._domain.topological_dimension()
        if t < 2:
            raise ValueError("CellFacetJacobian is only defined for topological dimensions >= 2.")

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        t = self._domain.topological_dimension()
        return (t, t - 1)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # This is always a constant mapping between two reference
        # coordinate systems.
        return True


@ufl_type()
class ReferenceCellEdgeVectors(GeometricCellQuantity):
    """The vectors between reference cell vertices for each edge in cell."""

    __slots__ = ()
    name = "RCEV"

    def __init__(self, domain):
        """Initialise."""
        GeometricCellQuantity.__init__(self, domain)
        t = self._domain.topological_dimension()
        if t < 2:
            raise ValueError("CellEdgeVectors is only defined for topological dimensions >= 2.")

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        cell = extract_unique_domain(self).ufl_cell()
        ne = cell.num_edges()
        t = cell.topological_dimension()
        return (ne, t)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # This is always constant for a given cell type
        return True


@ufl_type()
class ReferenceFacetEdgeVectors(GeometricFacetQuantity):
    """The vectors between reference cell vertices for each edge in current facet."""

    __slots__ = ()
    name = "RFEV"

    def __init__(self, domain):
        """Initialise."""
        GeometricFacetQuantity.__init__(self, domain)
        t = self._domain.topological_dimension()
        if t < 3:
            raise ValueError("FacetEdgeVectors is only defined for topological dimensions >= 3.")

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        cell = extract_unique_domain(self).ufl_cell()
        facet_types = cell.facet_types()

        # Raise exception for cells with more than one facet type e.g. prisms
        if len(facet_types) > 1:
            raise Exception(f"Cell type {cell} not supported.")

        nfe = facet_types[0].num_edges()
        t = cell.topological_dimension()
        return (nfe, t)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # This is always constant for a given cell type
        return True


@ufl_type()
class CellVertices(GeometricCellQuantity):
    """Physical cell vertices."""

    __slots__ = ()
    name = "CV"

    def __init__(self, domain):
        """Initialise."""
        GeometricCellQuantity.__init__(self, domain)

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        domain = extract_unique_domain(self)
        cell = domain.ufl_cell()
        nv = cell.num_vertices()
        g = domain.geometric_dimension()
        return (nv, g)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # This is always constant for a given cell type
        return True


@ufl_type()
class CellEdgeVectors(GeometricCellQuantity):
    """The vectors between physical cell vertices for each edge in cell."""

    __slots__ = ()
    name = "CEV"

    def __init__(self, domain):
        """Initialise."""
        GeometricCellQuantity.__init__(self, domain)
        t = self._domain.topological_dimension()
        if t < 2:
            raise ValueError("CellEdgeVectors is only defined for topological dimensions >= 2.")

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        domain = extract_unique_domain(self)
        cell = domain.ufl_cell()
        ne = cell.num_edges()
        g = domain.geometric_dimension()
        return (ne, g)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # This is always constant for a given cell type
        return True


@ufl_type()
class FacetEdgeVectors(GeometricFacetQuantity):
    """The vectors between physical cell vertices for each edge in current facet."""

    __slots__ = ()
    name = "FEV"

    def __init__(self, domain):
        """Initialise."""
        GeometricFacetQuantity.__init__(self, domain)
        t = self._domain.topological_dimension()
        if t < 3:
            raise ValueError("FacetEdgeVectors is only defined for topological dimensions >= 3.")

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        domain = extract_unique_domain(self)
        cell = domain.ufl_cell()
        facet_types = cell.facet_types()

        # Raise exception for cells with more than one facet type e.g. prisms
        if len(facet_types) > 1:
            raise Exception(f"Cell type {cell} not supported.")

        nfe = facet_types[0].num_edges()
        g = domain.geometric_dimension()
        return (nfe, g)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # This is always constant for a given cell type
        return True


# --- Determinants (signed or pseudo) of geometry mapping Jacobians


@ufl_type()
class JacobianDeterminant(GeometricCellQuantity):
    """The determinant of the Jacobian.

    Represents the signed determinant of a square Jacobian or the
    pseudo-determinant of a non-square Jacobian.
    """

    __slots__ = ()
    name = "detJ"

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # Only true for a piecewise linear coordinate field in simplex
        # cells
        return self._domain.is_piecewise_linear_simplex_domain()


@ufl_type()
class FacetJacobianDeterminant(GeometricFacetQuantity):
    """The pseudo-determinant of the FacetJacobian."""

    __slots__ = ()
    name = "detFJ"

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # Only true for a piecewise linear coordinate field in simplex cells
        return self._domain.is_piecewise_linear_simplex_domain()


@ufl_type()
class CellFacetJacobianDeterminant(GeometricFacetQuantity):
    """The pseudo-determinant of the CellFacetJacobian."""

    __slots__ = ()
    name = "detCFJ"

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # Only true for a piecewise linear coordinate field in simplex
        # cells
        return self._domain.is_piecewise_linear_simplex_domain()


# --- Inverses (signed or pseudo) of geometry mapping Jacobians


@ufl_type()
class JacobianInverse(GeometricCellQuantity):
    """The inverse of the Jacobian.

    Represents the inverse of a square Jacobian or the pseudo-inverse of
    a non-square Jacobian.
    """

    __slots__ = ()
    name = "K"

    @property
    def ufl_shape(self):
        """Return the number of coordinates defined (i.e. the geometric dimension of the domain)."""
        g = self._domain.geometric_dimension()
        t = self._domain.topological_dimension()
        return (t, g)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # Only true for a piecewise linear coordinate field in simplex
        # cells
        return self._domain.is_piecewise_linear_simplex_domain()


@ufl_type()
class FacetJacobianInverse(GeometricFacetQuantity):
    """The pseudo-inverse of the FacetJacobian."""

    __slots__ = ()
    name = "FK"

    def __init__(self, domain):
        """Initialise."""
        GeometricFacetQuantity.__init__(self, domain)
        t = self._domain.topological_dimension()
        if t < 2:
            raise ValueError(
                "FacetJacobianInverse is only defined for topological dimensions >= 2."
            )

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        g = self._domain.geometric_dimension()
        t = self._domain.topological_dimension()
        return (t - 1, g)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # Only true for a piecewise linear coordinate field in simplex
        # cells
        return self._domain.is_piecewise_linear_simplex_domain()


@ufl_type()
class CellFacetJacobianInverse(GeometricFacetQuantity):
    """The pseudo-inverse of the CellFacetJacobian."""

    __slots__ = ()
    name = "CFK"

    def __init__(self, domain):
        """Initialise."""
        GeometricFacetQuantity.__init__(self, domain)
        t = self._domain.topological_dimension()
        if t < 2:
            raise ValueError(
                "CellFacetJacobianInverse is only defined for topological dimensions >= 2."
            )

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        t = self._domain.topological_dimension()
        return (t - 1, t)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # Only true for a piecewise linear coordinate field in simplex cells
        return self._domain.is_piecewise_linear_simplex_domain()


# --- Types representing normal or tangent vectors


@ufl_type()
class FacetNormal(GeometricFacetQuantity):
    """The outwards pointing normal vector of the current facet."""

    __slots__ = ()
    name = "n"

    @property
    def ufl_shape(self):
        """Return the number of coordinates defined (i.e. the geometric dimension of the domain)."""
        g = self._domain.geometric_dimension()
        return (g,)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # For product cells, this is only true for some but not all
        # facets. Seems like too much work to fix right now.  Only
        # true for a piecewise linear coordinate field with simplex
        # _facets_.
        ce = self._domain.ufl_coordinate_element()
        is_piecewise_linear = ce.embedded_superdegree <= 1 and ce in H1
        return is_piecewise_linear and self._domain.ufl_cell().has_simplex_facets()


@ufl_type()
class CellNormal(GeometricCellQuantity):
    """The upwards pointing normal vector of the current manifold cell."""

    __slots__ = ()
    name = "cell_normal"

    @property
    def ufl_shape(self):
        """Return the number of coordinates defined (i.e. the geometric dimension of the domain)."""
        g = self._domain.geometric_dimension()
        # t = self._domain.topological_dimension()
        # return (g-t,g) # TODO: Should it be CellNormals? For interval in 3D we have two!
        return (g,)

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # Only true for a piecewise linear coordinate field in simplex cells
        return self._domain.is_piecewise_linear_simplex_domain()


@ufl_type()
class ReferenceNormal(GeometricFacetQuantity):
    """The outwards pointing normal vector of the current facet on the reference cell."""

    __slots__ = ()
    name = "reference_normal"

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        t = self._domain.topological_dimension()
        return (t,)


# --- Types representing measures of the cell and entities of the cell,
# typically used for stabilisation terms

# TODO: Clean up this set of types? Document!


@ufl_type()
class ReferenceCellVolume(GeometricCellQuantity):
    """The volume of the reference cell."""

    __slots__ = ()
    name = "reference_cell_volume"


@ufl_type()
class ReferenceFacetVolume(GeometricFacetQuantity):
    """The volume of the reference cell of the current facet."""

    __slots__ = ()
    name = "reference_facet_volume"


@ufl_type()
class CellVolume(GeometricCellQuantity):
    """The volume of the cell."""

    __slots__ = ()
    name = "volume"


@ufl_type()
class Circumradius(GeometricCellQuantity):
    """The circumradius of the cell."""

    __slots__ = ()
    name = "circumradius"


@ufl_type()
class CellDiameter(GeometricCellQuantity):
    """The diameter of the cell, i.e., maximal distance of two points in the cell."""

    __slots__ = ()
    name = "diameter"


@ufl_type()
class FacetArea(GeometricFacetQuantity):  # FIXME: Should this be allowed for interval domain?
    """The area of the facet."""

    __slots__ = ()
    name = "facetarea"


@ufl_type()
class MinCellEdgeLength(GeometricCellQuantity):
    """The minimum edge length of the cell."""

    __slots__ = ()
    name = "mincelledgelength"


@ufl_type()
class MaxCellEdgeLength(GeometricCellQuantity):
    """The maximum edge length of the cell."""

    __slots__ = ()
    name = "maxcelledgelength"


@ufl_type()
class MinFacetEdgeLength(GeometricFacetQuantity):
    """The minimum edge length of the facet."""

    __slots__ = ()
    name = "minfacetedgelength"


@ufl_type()
class MaxFacetEdgeLength(GeometricFacetQuantity):
    """The maximum edge length of the facet."""

    __slots__ = ()
    name = "maxfacetedgelength"


# --- Types representing other stuff


@ufl_type()
class CellOrientation(GeometricCellQuantity):
    """The orientation (+1/-1) of the current cell.

    For non-manifold cells (tdim == gdim), this equals the sign
    of the Jacobian determinant, i.e. +1 if the physical cell is
    oriented the same way as the reference cell and -1 otherwise.

    For manifold cells of tdim==gdim-1 this is input data belonging
    to the mesh, used to distinguish between the sides of the manifold.
    """

    __slots__ = ()
    name = "cell_orientation"


@ufl_type()
class FacetOrientation(GeometricFacetQuantity):
    """The orientation (+1/-1) of the current facet relative to the reference cell."""

    __slots__ = ()
    name = "facet_orientation"


# This doesn't quite fit anywhere. Make a special set of symbolic
# terminal types instead?
@ufl_type()
class QuadratureWeight(GeometricQuantity):
    """The current quadrature weight.

    Only used inside a quadrature context.
    """

    __slots__ = ()
    name = "weight"

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # The weight usually varies with the quadrature points
        return False



"""Utilites for types with a globally counted unique id attached to each object."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Massimiliano Leoni, 2016


def attach_ufl_id(cls):
    """Equip class with ``.ufl_id()`` and handle bookkeeping.

    Usage:

        1. Apply to class::

            @attach_ufl_id
            class MyClass(object):

        2. If ``__slots__`` is defined, include ``_ufl_id`` attribute::

            __slots__ = ("_ufl_id",)

        3. Add keyword argument to constructor::

            def __init__(self, *args, ufl_id=None):

        4. Call ``self._init_ufl_id`` with ``ufl_id`` and assign to ``._ufl_id``
           attribute::

            self._ufl_id = self._init_ufl_id(ufl_id)

    Result:

        ``MyClass().ufl_id()`` returns unique value for each constructed object.

    """

    def _get_ufl_id(self):
        """Return the ufl_id of this object."""
        return self._ufl_id

    def _init_ufl_id(cls):
        """Initialize new ufl_id for the object under construction."""
        # Bind cls with closure here

        def init_ufl_id(self, ufl_id):
            if ufl_id is None:
                ufl_id = cls._ufl_global_id
            cls._ufl_global_id = max(ufl_id, cls._ufl_global_id) + 1
            return ufl_id

        return init_ufl_id

    # Modify class:
    cls._ufl_global_id = 0
    cls.ufl_id = _get_ufl_id
    cls._init_ufl_id = _init_ufl_id(cls)
    return cls



__authors__ = "Martin Sandve Alnæs"
__date__ = "2008-03-12 -- 2009-01-28"

# Modified by Anders Logg, 2008
# Modified by Garth N. Wells, 2009

import pytest

from ufl import (
    Argument,
    Coefficient,
    FacetNormal,
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    adjoint,
    div,
    dot,
    ds,
    dx,
    grad,
    inner,
    triangle,
)
from ufl.algorithms import (
    expand_derivatives,
    expand_indices,
    extract_arguments,
    extract_coefficients,
    extract_elements,
    extract_unique_elements,
)
from ufl.corealg.traversal import (
    post_traversal,
    pre_traversal,
    unique_post_traversal,
    unique_pre_traversal,
)
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

# TODO: add more tests, covering all utility algorithms


@pytest.fixture(scope="module")
def element():
    return FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)


@pytest.fixture(scope="module")
def domain():
    return Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))


@pytest.fixture(scope="module")
def space(element, domain):
    return FunctionSpace(domain, element)


@pytest.fixture(scope="module")
def arguments(space):
    v = TestFunction(space)
    u = TrialFunction(space)
    return (v, u)


@pytest.fixture(scope="module")
def coefficients(space):
    c = Coefficient(space)
    f = Coefficient(space)
    return (c, f)


@pytest.fixture
def forms(arguments, coefficients, domain):
    v, u = arguments
    c, f = coefficients
    n = FacetNormal(domain)
    a = u * v * dx
    L = f * v * dx
    b = u * v * dx(0) + inner(c * grad(u), grad(v)) * dx(1) + dot(n, grad(u)) * v * ds + f * v * dx
    return (a, L, b)


def test_extract_arguments_vs_fixture(arguments, forms):
    assert arguments == tuple(extract_arguments(forms[0]))
    assert tuple(arguments[:1]) == tuple(extract_arguments(forms[1]))


def test_extract_coefficients_vs_fixture(coefficients, forms):
    assert coefficients == tuple(extract_coefficients(forms[2]))


def test_extract_elements_and_extract_unique_elements(forms, element, domain):
    b = forms[2]
    integrals = b.integrals_by_type("cell")
    integrals[0].integrand()

    element1 = element
    element2 = element

    space1 = FunctionSpace(domain, element1)
    space2 = FunctionSpace(domain, element2)

    v = TestFunction(space1)
    u = TrialFunction(space2)

    a = u * v * dx
    assert extract_elements(a) == (element1, element2)
    assert extract_unique_elements(a) == (element1,)


def test_pre_and_post_traversal(space):
    v = TestFunction(space)
    f = Coefficient(space)
    g = Coefficient(space)
    p1 = f * v
    p2 = g * v
    s = p1 + p2

    # NB! These traversal algorithms are intended to guarantee only
    # parent before child and vice versa, not this particular
    # ordering:
    assert list(pre_traversal(s)) == [s, p2, g, v, p1, f, v]
    assert list(post_traversal(s)) == [g, v, p2, f, v, p1, s]
    assert list(unique_pre_traversal(s)) == [s, p2, g, v, p1, f]
    assert list(unique_post_traversal(s)) == [v, f, p1, g, p2, s]


def test_expand_indices(domain):
    element = FiniteElement("Lagrange", triangle, 2, (), identity_pullback, H1)
    space = FunctionSpace(domain, element)
    v = TestFunction(space)
    u = TrialFunction(space)

    def evaluate(form):
        return form.cell_integral()[0].integrand()(
            (), {v: 3, u: 5}
        )  # TODO: How to define values of derivatives?

    a = div(grad(v)) * u * dx
    # a1 = evaluate(a)
    a = expand_derivatives(a)
    # a2 = evaluate(a)
    a = expand_indices(a)
    # a3 = evaluate(a)
    # TODO: Compare a1, a2, a3
    # TODO: Test something more


def test_adjoint(domain):
    cell = triangle

    V1 = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    V2 = FiniteElement("Lagrange", cell, 2, (), identity_pullback, H1)

    s1 = FunctionSpace(domain, V1)
    s2 = FunctionSpace(domain, V2)

    u = TrialFunction(s1)
    v = TestFunction(s2)
    assert u.number() > v.number()

    u2 = Argument(s1, 2)
    v2 = Argument(s2, 3)
    assert u2.number() < v2.number()

    a = u * v * dx
    a_arg_degrees = [arg.ufl_element().embedded_superdegree for arg in extract_arguments(a)]
    assert a_arg_degrees == [2, 1]

    b = adjoint(a)
    b_arg_degrees = [arg.ufl_element().embedded_superdegree for arg in extract_arguments(b)]
    assert b_arg_degrees == [1, 2]

    c = adjoint(a, (u2, v2))
    c_arg_degrees = [arg.ufl_element().embedded_superdegree for arg in extract_arguments(c)]
    assert c_arg_degrees == [1, 2]

    d = adjoint(b)
    d_arg_degrees = [arg.ufl_element().embedded_superdegree for arg in extract_arguments(d)]
    assert d_arg_degrees == [2, 1]



from ufl import Cell, Mesh
from ufl.finiteelement import FiniteElement, MixedElement, SymmetricElement
from ufl.functionspace import FunctionSpace
from ufl.pullback import contravariant_piola, covariant_piola, identity_pullback
from ufl.sobolevspace import H1, HCurl, HDiv


def test_reference_shapes():
    # show_elements()
    cell = Cell("triangle")
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (3,), identity_pullback, H1))

    V = FiniteElement("N1curl", cell, 1, (2,), covariant_piola, HCurl)
    Vspace = FunctionSpace(domain, V)
    assert Vspace.value_shape == (3,)
    assert V.reference_value_shape == (2,)

    U = FiniteElement("Raviart-Thomas", cell, 1, (2,), contravariant_piola, HDiv)
    Uspace = FunctionSpace(domain, U)
    assert Uspace.value_shape == (3,)
    assert U.reference_value_shape == (2,)

    W = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    Wspace = FunctionSpace(domain, W)
    assert Wspace.value_shape == ()
    assert W.reference_value_shape == ()

    Q = FiniteElement("Lagrange", cell, 1, (3,), identity_pullback, H1)
    Qspace = FunctionSpace(domain, Q)
    assert Qspace.value_shape == (3,)
    assert Q.reference_value_shape == (3,)

    T = FiniteElement("Lagrange", cell, 1, (3, 3), identity_pullback, H1)
    Tspace = FunctionSpace(domain, T)
    assert Tspace.value_shape == (3, 3)
    assert T.reference_value_shape == (3, 3)

    S = SymmetricElement(
        {
            (0, 0): 0,
            (1, 0): 1,
            (2, 0): 2,
            (0, 1): 1,
            (1, 1): 3,
            (2, 1): 4,
            (0, 2): 2,
            (1, 2): 4,
            (2, 2): 5,
        },
        [FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1) for _ in range(6)],
    )
    Sspace = FunctionSpace(domain, S)
    assert Sspace.value_shape == (3, 3)
    assert S.reference_value_shape == (6,)

    M = MixedElement([V, U, W])
    Mspace = FunctionSpace(domain, M)
    assert Mspace.value_shape == (7,)
    assert M.reference_value_shape == (5,)



"""Algorithm to check for 'comparison' nodes in a form when the user is in 'complex mode'."""

from ufl.algebra import Real
from ufl.algorithms.map_integrands import map_integrand_dags
from ufl.argument import Argument
from ufl.constantvalue import RealValue, Zero
from ufl.corealg.multifunction import MultiFunction
from ufl.geometry import GeometricQuantity


class CheckComparisons(MultiFunction):
    """Raises an error if comparisons are done with complex quantities.

    If quantities are real, adds the Real operator to the compared quantities.

    Terminals that are real are RealValue, Zero, and Argument
    (even in complex FEM, the basis functions are real)
    Operations that produce reals are Abs, Real, Imag.
    Terminals default to complex, and Sqrt, Pow (defensively) imply complex.
    Otherwise, operators preserve the type of their operands.
    """

    def __init__(self):
        """Initialise."""
        MultiFunction.__init__(self)
        self.nodetype = {}

    def expr(self, o, *ops):
        """Defaults expressions to complex unless they only act on real quantities.

        Overridden for specific operators. Rebuilds objects if necessary.
        """
        types = {self.nodetype[op] for op in ops}

        if types:
            t = "complex" if "complex" in types else "real"
        else:
            t = "complex"

        o = self.reuse_if_untouched(o, *ops)
        self.nodetype[o] = t
        return o

    def compare(self, o, *ops):
        """Compare."""
        types = {self.nodetype[op] for op in ops}

        if "complex" in types:
            raise ComplexComparisonError("Ordering undefined for complex values.")
        else:
            o = o._ufl_expr_reconstruct_(*map(Real, ops))
            self.nodetype[o] = "bool"
            return o

    gt = compare
    lt = compare
    ge = compare
    le = compare
    sign = compare

    def max_value(self, o, *ops):
        """Apply to max_value."""
        types = {self.nodetype[op] for op in ops}

        if "complex" in types:
            raise ComplexComparisonError("You can't compare complex numbers with max.")
        else:
            o = o._ufl_expr_reconstruct_(*map(Real, ops))
            self.nodetype[o] = "bool"
            return o

    def min_value(self, o, *ops):
        """Apply to min_value."""
        types = {self.nodetype[op] for op in ops}

        if "complex" in types:
            raise ComplexComparisonError("You can't compare complex numbers with min.")
        else:
            o = o._ufl_expr_reconstruct_(*map(Real, ops))
            self.nodetype[o] = "bool"
            return o

    def real(self, o, *ops):
        """Apply to real."""
        o = self.reuse_if_untouched(o, *ops)
        self.nodetype[o] = "real"
        return o

    def imag(self, o, *ops):
        """Apply to imag."""
        o = self.reuse_if_untouched(o, *ops)
        self.nodetype[o] = "real"
        return o

    def sqrt(self, o, *ops):
        """Apply to sqrt."""
        o = self.reuse_if_untouched(o, *ops)
        self.nodetype[o] = "complex"
        return o

    def power(self, o, base, exponent):
        """Apply to power."""
        o = self.reuse_if_untouched(o, base, exponent)
        try:
            # Attempt to diagnose circumstances in which the result must be real.
            exponent = float(exponent)
            if self.nodetype[base] == "real" and int(exponent) == exponent:
                self.nodetype[o] = "real"
                return o
        except TypeError:
            pass

        self.nodetype[o] = "complex"
        return o

    def abs(self, o, *ops):
        """Apply to abs."""
        o = self.reuse_if_untouched(o, *ops)
        self.nodetype[o] = "real"
        return o

    def terminal(self, term, *ops):
        """Apply to terminal."""
        # default terminals to complex, except the ones we *know* are real
        if isinstance(term, (RealValue, Zero, Argument, GeometricQuantity)):
            self.nodetype[term] = "real"
        else:
            self.nodetype[term] = "complex"
        return term

    def indexed(self, o, expr, multiindex):
        """Apply to indexed."""
        o = self.reuse_if_untouched(o, expr, multiindex)
        self.nodetype[o] = self.nodetype[expr]
        return o


def do_comparison_check(form):
    """Raises an error if invalid comparison nodes exist."""
    return map_integrand_dags(CheckComparisons(), form)


class ComplexComparisonError(BaseException):
    """Complex compariseon exception."""



import pytest

from ufl import Argument, Coefficient, FunctionSpace, Mesh, triangle
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

# TODO: Add more illegal expressions to check!


@pytest.fixture
def selement():
    return FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)


@pytest.fixture
def velement():
    return FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)


@pytest.fixture
def domain():
    return Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))


@pytest.fixture
def sspace(domain, selement):
    return FunctionSpace(domain, selement)


@pytest.fixture
def vspace(domain, velement):
    return FunctionSpace(domain, velement)


@pytest.fixture
def a(sspace):
    return Argument(sspace, 2)


@pytest.fixture
def b(sspace):
    return Argument(sspace, 3)


@pytest.fixture
def v(vspace):
    return Argument(vspace, 4)


@pytest.fixture
def u(vspace):
    return Argument(vspace, 5)


@pytest.fixture
def f(sspace):
    return Coefficient(sspace)


@pytest.fixture
def g(sspace):
    return Coefficient(sspace)


@pytest.fixture
def vf(vspace):
    return Coefficient(vspace)


@pytest.fixture
def vg(vspace):
    return Coefficient(vspace)


def test_mul_v_u(v, u):
    with pytest.raises(BaseException):
        v * u


def test_mul_vf_u(vf, u):
    with pytest.raises(BaseException):
        vf * u


def test_mul_vf_vg(vf, vg):
    with pytest.raises(BaseException):
        vf * vg


def test_add_a_v(a, v):
    with pytest.raises(BaseException):
        a + v


def test_add_vf_b(vf, b):
    with pytest.raises(BaseException):
        vf + b


def test_add_vectorexpr_b(vg, v, u, vf, b):
    tmp = vg + v + u + vf
    with pytest.raises(BaseException):
        tmp + b



"""This module defines expression transformation utilities.

These utilities are for expanding free indices in expressions to explicit fixed indices only.
"""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2009.

from ufl.algorithms.transformer import ReuseTransformer, apply_transformer
from ufl.classes import Terminal
from ufl.constantvalue import Zero
from ufl.core.multiindex import FixedIndex, Index, MultiIndex
from ufl.differentiation import Grad
from ufl.utils.stacks import Stack, StackDict


class IndexExpander(ReuseTransformer):
    """Index expander."""

    def __init__(self):
        """Initialise."""
        ReuseTransformer.__init__(self)
        self._components = Stack()
        self._index2value = StackDict()

    def component(self):
        """Return current component tuple."""
        if self._components:
            return self._components.peek()
        return ()

    def terminal(self, x):
        """Apply to terminal."""
        if x.ufl_shape:
            c = self.component()
            if len(x.ufl_shape) != len(c):
                raise ValueError("Component size mismatch.")
            return x[c]
        return x

    def form_argument(self, x):
        """Apply to form_argument."""
        sh = x.ufl_shape
        if sh == ():
            return x
        else:
            space = x.ufl_function_space()
            r = len(sh)

            # Get component
            c = self.component()
            if r != len(c):
                raise ValueError("Component size mismatch.")

            # Map it through an eventual symmetry mapping
            if len(space.components) > 1:
                c = min(i for i, j in space.components.items() if j == space.components[c])
                if r != len(c):
                    raise ValueError("Component size mismatch after symmetry mapping.")

            return x[c]

    def zero(self, x):
        """Apply to zero."""
        if len(x.ufl_shape) != len(self.component()):
            raise ValueError("Component size mismatch.")

        s = set(x.ufl_free_indices) - set(i.count() for i in self._index2value.keys())
        if s:
            raise ValueError(f"Free index set mismatch, these indices have no value assigned: {s}.")

        # There is no index/shape info in this zero because that is asserted above
        return Zero()

    def scalar_value(self, x):
        """Apply to scalar_value."""
        if len(x.ufl_shape) != len(self.component()):
            self.print_visit_stack()
        if len(x.ufl_shape) != len(self.component()):
            raise ValueError("Component size mismatch.")

        s = set(x.ufl_free_indices) - set(i.count() for i in self._index2value.keys())
        if s:
            raise ValueError(f"Free index set mismatch, these indices have no value assigned: {s}.")

        return x._ufl_class_(x.value())

    def conditional(self, x):
        """Apply to conditional."""
        c, t, f = x.ufl_operands

        # Not accepting nonscalars in condition
        if c.ufl_shape != ():
            raise ValueError("Not expecting tensor in condition.")

        # Conditional may be indexed, push empty component
        self._components.push(())
        c = self.visit(c)
        self._components.pop()

        # Keep possibly non-scalar components for values
        t = self.visit(t)
        f = self.visit(f)

        return self.reuse_if_possible(x, c, t, f)

    def division(self, x):
        """Apply to division."""
        a, b = x.ufl_operands

        # Not accepting nonscalars in division anymore
        if a.ufl_shape != ():
            raise ValueError("Not expecting tensor in division.")
        if self.component() != ():
            raise ValueError("Not expecting component in division.")

        if b.ufl_shape != ():
            raise ValueError("Not expecting division by tensor.")
        a = self.visit(a)

        # self._components.push(())
        b = self.visit(b)
        # self._components.pop()

        return self.reuse_if_possible(x, a, b)

    def index_sum(self, x):
        """Apply to index_sum."""
        ops = []
        summand, multiindex = x.ufl_operands
        (index,) = multiindex

        # TODO: For the list tensor purging algorithm, do something like:
        # if index not in self._to_expand:
        #     return self.expr(x, *[self.visit(o) for o in x.ufl_operands])

        for value in range(x.dimension()):
            self._index2value.push(index, value)
            ops.append(self.visit(summand))
            self._index2value.pop()
        return sum(ops)

    def _multi_index_values(self, x):
        """Apply to _multi_index_values."""
        comp = []
        for i in x._indices:
            if isinstance(i, FixedIndex):
                comp.append(i._value)
            elif isinstance(i, Index):
                comp.append(self._index2value[i])
        return tuple(comp)

    def multi_index(self, x):
        """Apply to multi_index."""
        comp = self._multi_index_values(x)
        return MultiIndex(tuple(FixedIndex(i) for i in comp))

    def indexed(self, x):
        """Apply to indexed."""
        A, ii = x.ufl_operands

        # Push new component built from index value map
        self._components.push(self._multi_index_values(ii))

        # Hide index values (doing this is not correct behaviour)
        # for i in ii:
        #     if isinstance(i, Index):
        #         self._index2value.push(i, None)

        result = self.visit(A)

        # Un-hide index values
        # for i in ii:
        #     if isinstance(i, Index):
        #         self._index2value.pop()

        # Reset component
        self._components.pop()
        return result

    def component_tensor(self, x):
        """Apply to component_tensor."""
        # This function evaluates the tensor expression
        # with indices equal to the current component tuple
        expression, indices = x.ufl_operands
        if expression.ufl_shape != ():
            raise ValueError("Expecting scalar base expression.")

        # Update index map with component tuple values
        comp = self.component()
        if len(indices) != len(comp):
            raise ValueError("Index/component mismatch.")
        for i, v in zip(indices.indices(), comp):
            self._index2value.push(i, v)
        self._components.push(())

        # Evaluate with these indices
        result = self.visit(expression)

        # Revert index map
        for _ in comp:
            self._index2value.pop()
        self._components.pop()
        return result

    def list_tensor(self, x):
        """Apply to list_tensor."""
        # Pick the right subtensor and subcomponent
        c = self.component()
        c0, c1 = c[0], c[1:]
        op = x.ufl_operands[c0]
        # Evaluate subtensor with this subcomponent
        self._components.push(c1)
        r = self.visit(op)
        self._components.pop()
        return r

    def grad(self, x):
        """Apply to grad."""
        (f,) = x.ufl_operands
        if not isinstance(f, (Terminal, Grad)):
            raise ValueError("Expecting expand_derivatives to have been applied.")
        # No need to visit child as long as it is on the form [Grad]([Grad](terminal))
        return x[self.component()]


def expand_indices(e):
    """Expand indices."""
    return apply_transformer(e, IndexExpander())



"""Collection of utilities for mapping between multiindices and a flattened index space."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later


def shape_to_strides(sh):
    """Return a tuple of strides given a shape tuple."""
    n = len(sh)
    if not n:
        return ()
    strides = [None] * n
    strides[n - 1] = 1
    for i in range(n - 1, 0, -1):
        strides[i - 1] = strides[i] * sh[i]
    return tuple(strides)


def flatten_multiindex(ii, strides):
    """Return the flat index corresponding to the given multiindex."""
    i = 0
    for c, s in zip(ii, strides):
        i += c * s
    return i


def unflatten_index(i, strides):
    """Return the multiindex corresponding to the given flat index."""
    ii = []
    for s in strides:
        ii.append(i // s)
        i %= s
    return tuple(ii)



"""Utility functions for checking properties of expressions."""

# Copyright (C) 2013-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2008-2009

from ufl.core.expr import Expr
from ufl.core.terminal import FormArgument
from ufl.corealg.traversal import traverse_unique_terminals
from ufl.geometry import GeometricQuantity
from ufl.sobolevspace import H1


def is_python_scalar(expression):
    """Return True iff expression is of a Python scalar type."""
    return isinstance(expression, (int, float, complex))


def is_ufl_scalar(expression):
    """Return True iff expression is scalar-valued, but possibly containing free indices."""
    return isinstance(expression, Expr) and not expression.ufl_shape


def is_true_ufl_scalar(expression):
    """Return True iff expression is scalar-valued, with no free indices."""
    return isinstance(expression, Expr) and not (
        expression.ufl_shape or expression.ufl_free_indices
    )


def is_cellwise_constant(expr):
    """Return whether expression is constant over a single cell."""
    # TODO: Implement more accurately considering e.g. derivatives?
    return all(e.is_cellwise_constant() for e in traverse_unique_terminals(expr))


def is_scalar_constant_expression(expr):
    """Check if an expression is a globally constant scalar expression."""
    if is_python_scalar(expr):
        return True
    if expr.ufl_shape:
        return False

    # TODO: This does not consider gradients of coefficients, so false
    # negatives are possible.
    for e in traverse_unique_terminals(expr):
        # Return False if any single terminal is not constant
        if isinstance(e, FormArgument):
            # Accept only globally constant Arguments and Coefficients
            if e.ufl_element().embedded_superdegree > 0 or e.ufl_element() not in H1:
                return False
        elif isinstance(e, GeometricQuantity):
            # Reject all geometric quantities, they all vary over
            # cells
            return False

    # All terminals passed constant check
    return True



#
# Author: Martin Sandve Alnes
# Date: 2008-10-03
#
from ufl import (
    Coefficient,
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    dot,
    dx,
    grad,
    triangle,
)
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
space = FunctionSpace(domain, element)

u = TrialFunction(space)
v = TestFunction(space)
w = Coefficient(space)

a = dot(dot(w, grad(u)), v) * dx



"""Classes.

This file is useful for external code like tests and form compilers,
since it enables the syntax "from ufl.classes import CellFacetooBar" for getting
implementation details not exposed through the default ufl namespace.
It also contains functionality used by algorithms for dealing with groups
of classes, and for mapping types to different handler functions.
"""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2009.
# Modified by Kristian B. Oelgaard, 2011
# Modified by Andrew T. T. McRae, 2014

# This will be populated part by part below
__all__ = []

# Import all submodules, triggering execution of the ufl_type class
# decorator for each Expr class.

import ufl.algebra
import ufl.argument
import ufl.averaging
import ufl.cell
import ufl.coefficient
import ufl.conditional
import ufl.constantvalue
import ufl.core.expr
import ufl.core.multiindex
import ufl.core.operator
import ufl.core.terminal
import ufl.differentiation
import ufl.domain
import ufl.equation
import ufl.exprcontainers
import ufl.finiteelement
import ufl.form
import ufl.functionspace
import ufl.geometry
import ufl.indexed
import ufl.indexsum
import ufl.integral
import ufl.mathfunctions
import ufl.measure
import ufl.pullback
import ufl.referencevalue
import ufl.restriction
import ufl.sobolevspace
import ufl.tensoralgebra
import ufl.tensors
import ufl.variable
from ufl import exproperators as __exproperators

# Collect all classes in sets automatically classified by some properties
all_ufl_classes = set(ufl.core.expr.Expr._ufl_all_classes_)
abstract_classes = set(c for c in all_ufl_classes if c._ufl_is_abstract_)
ufl_classes = set(c for c in all_ufl_classes if not c._ufl_is_abstract_)
terminal_classes = set(c for c in all_ufl_classes if c._ufl_is_terminal_)
nonterminal_classes = set(c for c in all_ufl_classes if not c._ufl_is_terminal_)

__all__ += [
    "all_ufl_classes",
    "abstract_classes",
    "ufl_classes",
    "terminal_classes",
    "nonterminal_classes",
    "__exproperators",
]


def populate_namespace_with_expr_classes(namespace):
    """Export all Expr subclasses into the namespace under their natural name."""
    names = []
    classes = ufl.core.expr.Expr._ufl_all_classes_
    for cls in classes:
        class_name = cls.__name__
        namespace[class_name] = cls
        names.append(class_name)
    return names


__all__ += populate_namespace_with_expr_classes(locals())


# Semi-automated imports of non-expr classes:


def populate_namespace_with_module_classes(mod, loc):
    """Export the classes that submodules list in __all_classes__."""
    names = mod.__all_classes__
    for name in names:
        loc[name] = getattr(mod, name)
    return names


__all__ += populate_namespace_with_module_classes(ufl.cell, locals())
__all__ += populate_namespace_with_module_classes(ufl.finiteelement, locals())
__all__ += populate_namespace_with_module_classes(ufl.domain, locals())
__all__ += populate_namespace_with_module_classes(ufl.functionspace, locals())
__all__ += populate_namespace_with_module_classes(ufl.core.multiindex, locals())
__all__ += populate_namespace_with_module_classes(ufl.argument, locals())
__all__ += populate_namespace_with_module_classes(ufl.measure, locals())
__all__ += populate_namespace_with_module_classes(ufl.integral, locals())
__all__ += populate_namespace_with_module_classes(ufl.form, locals())
__all__ += populate_namespace_with_module_classes(ufl.equation, locals())
__all__ += populate_namespace_with_module_classes(ufl.pullback, locals())
__all__ += populate_namespace_with_module_classes(ufl.sobolevspace, locals())
